"""
__main__
~~~~~~~~

Mainline for the CLI interface of the :mod:`yadr` package.

.. autofunction:: yadr.__main__.parse_cli
"""
from argparse import ArgumentParser

from yadr.yadr import add_dice_map, list_dice_maps, roll


def parse_cli() -> None:
    """Parse command line options and execute those commands.

    :returns: `None`.
    :rtype: NoneType

    Invoking :mod:`yadr`
    --------------------
    After installing :mod:`yadr` with :mod:`pip` or similar process,
    you can run it from the command line with the following::

        $ yadr

    Options can be passed into :mod:`yadr` to alter its behavior. The
    easiest way to view these options is to invoke the help::

        $ yadr -h
    """
    # Stand up the parser.
    p = ArgumentParser(
        description='Execute YADN syntax to roll dice.',
        prog='yadr'
    )

    # Define the command line arguments.
    p.add_argument(
        'yadn',
        help='A string of YADN describing the die roll.',
        action='store',
        nargs='?',
        type=str
    )
    p.add_argument(
        '--list_dice_maps', '-l',
        help='List the names of the default dice maps.',
        action='store_true'
    )
    p.add_argument(
        '--add_dice_map', '-m',
        help='Load the dice mappings at the given file location.',
        nargs=1,
        action='store',
        type=str
    )

    # Parse and execute the command.
    args = p.parse_args()
    result = 'Use `yadr -h` to view the available options for running yadr.\n'
    dice_map = {}
    if args.add_dice_map:
        dice_map = add_dice_map(args.add_dice_map[0])
    elif args.list_dice_maps:
        result = list_dice_maps()
    if args.yadn:
        raw_result = roll(args.yadn, True, dice_map)
        result = str(raw_result)
    print(result)


if __name__ == '__main__':
    parse_cli()



"""
test_lex
~~~~~~~~

Unit tests for the dice notation lexer.
"""
from collections import namedtuple
from functools import partial

import pytest

from yadr import lex
from yadr import model as m


# Registration.
cases = {}


class reg:
    def __init__(self, key):
        self.key = key

    def __call__(self, fn):
        new = partial(fn, token=self.key)
        cases[self.key] = new
        return fn


# Common test case utilities.
YADN = namedtuple('YADN', 'yadn tokens')


def get_yadn(token):
    """Build a :class:`YADN` object for each valid symbol for the token."""
    # If a YADN object was passed, just yield it back.
    if isinstance(token, YADN):
        yield token

    # Otherwise get the symbols for the token type, build and then
    # yield the YADN object for each of the symbols.
    else:
        symbols = m.yadn_symbols_raw[token].split()
        for symbol in symbols:
            yadn = symbol
            value = symbol
            if token == m.Token.NUMBER:
                value = int(value)
            elif token == m.Token.BOOLEAN:
                value = True
                if symbol == 'F':
                    value = False
            yield YADN(yadn, ((token, value),))


# Build the test data for testing each type of token.
@reg(m.Token.CHOICE_OPERATOR)
def build_choice_operator_cases(token):
    yadns = get_yadn(token)
    for yadn in yadns:
        yield YADN(f'{yadn.yadn}"spam":"eggs"', (
            *yadn.tokens,
            (m.Token.QUALIFIER, 'spam'),
            (m.Token.OPTIONS_OPERATOR, ':'),
            (m.Token.QUALIFIER, 'eggs'),
        ))


@reg(m.Token.GROUP_OPEN)
def build_group_cases(token):
    yield YADN('(3+3)', (
        (m.Token.GROUP_OPEN, '('),
        (m.Token.NUMBER, 3),
        (m.Token.AS_OPERATOR, '+'),
        (m.Token.NUMBER, 3),
        (m.Token.GROUP_CLOSE, ')'),
    ))


@reg(m.Token.AS_OPERATOR)
@reg(m.Token.COMPARISON_OPERATOR)
@reg(m.Token.DICE_OPERATOR)
@reg(m.Token.EX_OPERATOR)
@reg(m.Token.MD_OPERATOR)
@reg(m.Token.POOL_DEGEN_OPERATOR)
@reg(m.Token.POOL_GEN_OPERATOR)
@reg(m.Token.POOL_OPERATOR)
@reg(m.Token.ROLL_DELIMITER)
def build_followed_by_number_cases(token):
    symbols = m.yadn_symbols_raw[token].split()
    for symbol in symbols:
        yield YADN(f'{symbol}3', (
            (token, symbol),
            (m.Token.NUMBER, 3),
        ))


@reg(m.Token.MAPPING_OPERATOR)
def build_mapping_operator_cases(token):
    for yadn in build_qualifier_cases():
        yield YADN(f'm{yadn.yadn}', (
            (m.Token.MAPPING_OPERATOR, 'm'),
            *yadn.tokens
        ))


@reg(m.Token.MAP)
@reg(m.Token.MAP_OPEN)
def build_map_cases(token):
    yield YADN('{"spam"=1:"win",2:"lose"}', ((m.Token.MAP, ('spam', {
        1: 'win',
        2: 'lose',
    })),))


@reg(m.Token.NEGATIVE_SIGN)
def build_negative_sign_cases(token):
    yield YADN('-2', ((m.Token.NUMBER, -2),))


@reg(m.Token.OPTIONS_OPERATOR)
def build_options_operator_cases(token):
    yield YADN(':"spam"', (
        (token, ':'),
        (m.Token.QUALIFIER, 'spam'),
    ))


@reg(m.Token.POOL)
@reg(m.Token.POOL_OPEN)
def build_pool_cases(token):
    yield YADN('[3,3]', ((m.Token.POOL, (3, 3)),))


@reg(m.Token.QUALIFIER)
@reg(m.Token.QUALIFIER_DELIMITER)
def build_qualifier_cases(token=None):
    yield YADN('"spam"', ((m.Token.QUALIFIER, 'spam'),))


@reg(m.Token.BOOLEAN)
@reg(m.Token.GROUP_CLOSE)
@reg(m.Token.NUMBER)
def build_token_only_case(token):
    for yadn in get_yadn(token):
        yield yadn


@reg(m.Token.U_POOL_DEGEN_OPERATOR)
def build_u_pool_degeneration_operator_cases(token):
    symbols = m.yadn_symbols_raw[m.Token.U_POOL_DEGEN_OPERATOR].split()
    for symbol in symbols:
        yadn = f'{symbol}[3, 3]'
        yield YADN(yadn, (
            (m.Token.U_POOL_DEGEN_OPERATOR, symbol),
            (m.Token.POOL, (3, 3)),
        ))


# Core test code.
def allowed_test(symbol, before, after):
    """Test a token that is allowed to follow the given symbol."""
    expected = (
        *before.tokens,
        *symbol.tokens,
        *after.tokens,
    )

    lexer = lex.Lexer()
    yadns = [
        f'{before.yadn}{symbol.yadn}{after.yadn}',
        f'{before.yadn}{symbol.yadn} {after.yadn}',
    ]

    for yadn in yadns:
        actual = lexer.lex(yadn)
        try:
            assert actual == expected
        except AssertionError as ex:
            cls = ex.__class__
            msg = f'{ex}\n{yadn}'
            raise cls(msg)


def disallowed_test(symbol, before, after):
    """Test a token that is not allowed to follow the given symbol"""
    expected = ValueError

    lexer = lex.Lexer()
    yadns = [
        f'{before.yadn}{symbol.yadn}{after.yadn}',
        f'{before.yadn}{symbol.yadn} {after.yadn}',
    ]

    for yadn in yadns:
        try:
            pytest.raises(expected, lexer.lex, yadn)
        except BaseException as ex:
            cls = ex.__class__
            msg = f'{ex}\n{yadn}'
            raise cls(msg)


def lexer_test(token, before, alloweds):
    """Common test of :meth:`Lexer.lex` for a token."""
    # Build list of disallowed tokens.
    disalloweds = [token for token in cases if token not in alloweds]

    # Since AS_OPERATORS contain -, the NEGATIVE_SIGN will be allowed
    # if AS_OPERATORS are allowed.
    if m.Token.AS_OPERATOR in alloweds:
        disalloweds = [
            token for token in disalloweds
            if token != m.Token.NEGATIVE_SIGN
        ]

    # Since AS_OPERATORS contain -, that AS_OPERATOR is allowed when
    # NEGATIVE_SIGN is allowed. This change does prevent us from testing
    # whether + is allowed, though.
    if m.Token.NEGATIVE_SIGN in alloweds:
        disalloweds = [
            token for token in disalloweds
            if token != m.Token.AS_OPERATOR
        ]

    # NUMBERS next to NUMBERS aren't separate tokens.
    if (
        token == m.Token.NUMBER
        or (
            isinstance(token, YADN)
            and token.tokens
            and token.tokens[0][0] == m.Token.NUMBER
        )
    ):
        disalloweds = [
            token for token in disalloweds
            if token != m.Token.NUMBER
        ]

    # Test each of the possible symbols for the token.
    symbols = get_yadn(token)
    for symbol in symbols:

        # Test alloweds.
        for key in alloweds:
            afters = cases[key]()
            for after in afters:
                allowed_test(symbol, before, after)

        # Test disalloweds.
        for key in disalloweds:
            afters = cases[key]()
            for after in afters:
                disallowed_test(symbol, before, after)


# Symbol unit tests.
def test_as_operator():
    """Given an addition or subtraction operator, return the correct
    tokens for the YADN.
    """
    token = m.Token.AS_OPERATOR
    before = YADN('3', ((m.Token.NUMBER, 3),))
    alloweds = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


def test_boolean():
    """Given an addition or subtraction operator, return the correct
    tokens for the YADN.
    """
    token = m.Token.BOOLEAN
    before = YADN('', ())
    alloweds = [
        m.Token.CHOICE_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


def test_choice_operator():
    """Given a choice, return the correct tokens for the YADN.
    """
    token = m.Token.CHOICE_OPERATOR
    before = YADN('T', ((m.Token.BOOLEAN, True),))
    alloweds = [
        m.Token.QUALIFIER,
        m.Token.QUALIFIER_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_comparison_operator():
    """Given a comparison operator, return the correct tokens for
    the YADN.
    """
    token = m.Token.COMPARISON_OPERATOR
    before = YADN('3', ((m.Token.NUMBER, 3),))
    alloweds = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


def test_dice_operator():
    """Given a comparison operator, return the correct tokens for
    the YADN.
    """
    token = m.Token.DICE_OPERATOR
    before = YADN('3', ((m.Token.NUMBER, 3),))
    alloweds = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


def test_ex_operator():
    """Given a exponentiation operator, return the correct tokens for
    the YADN.
    """
    token = m.Token.EX_OPERATOR
    before = YADN('3', ((m.Token.NUMBER, 3),))
    alloweds = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


def test_group():
    """Given a group close, return the correct tokens for the YADN."""
    token = m.Token.GROUP_CLOSE
    before = YADN('(3', (
        (m.Token.GROUP_OPEN, '('),
        (m.Token.NUMBER, 3),
    ))
    alloweds = [
        m.Token.AS_OPERATOR,
        m.Token.MD_OPERATOR,
        m.Token.EX_OPERATOR,
        m.Token.DICE_OPERATOR,
        m.Token.GROUP_CLOSE,
        m.Token.POOL_OPERATOR,
        m.Token.ROLL_DELIMITER,
        m.Token.POOL_GEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


def test_group_open():
    """Given a group open, return the correct tokens for the YADN."""
    token = m.Token.GROUP_OPEN
    before = YADN('', ())
    alloweds = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.POOL,
        m.Token.POOL_OPEN,
        m.Token.U_POOL_DEGEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


def test_map():
    """Given a map, return the correct tokens for the YADN."""
    token = YADN(
        '{"name"=1:"none",2:"success",3:"success",4:"success success"}',
        ((m.Token.MAP, ('name', {
            1: "none",
            2: "success",
            3: "success",
            4: "success success",
        })),)
    )
    before = YADN('', ())
    alloweds = [
        m.Token.ROLL_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_mapping_operator():
    """Given a mapping operator, return the correct tokens for the YADN."""
    token = m.Token.MAPPING_OPERATOR
    before = YADN('3', ((m.Token.NUMBER, 3),))
    alloweds = [
        m.Token.QUALIFIER,
        m.Token.QUALIFIER_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_number():
    """Given a number, return the correct tokens for the YADN."""
    token = m.Token.NUMBER
    before = YADN('', ())
    alloweds = [
        m.Token.AS_OPERATOR,
        m.Token.COMPARISON_OPERATOR,
        m.Token.EX_OPERATOR,
        m.Token.DICE_OPERATOR,
        m.Token.GROUP_CLOSE,
        m.Token.MAPPING_OPERATOR,
        m.Token.MD_OPERATOR,
        m.Token.POOL_GEN_OPERATOR,
        m.Token.ROLL_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_number_multiple_digits():
    """Given a number with multiple digits, return the correct tokens
    for the YADN.
    """
    token = YADN('34', ((m.Token.NUMBER, 34),))
    before = YADN('', ())
    alloweds = [
        m.Token.AS_OPERATOR,
        m.Token.COMPARISON_OPERATOR,
        m.Token.EX_OPERATOR,
        m.Token.DICE_OPERATOR,
        m.Token.GROUP_CLOSE,
        m.Token.MAPPING_OPERATOR,
        m.Token.MD_OPERATOR,
        m.Token.POOL_GEN_OPERATOR,
        m.Token.ROLL_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_options_operator():
    """Given a options operator, return the correct tokens for the YADN."""
    token = m.Token.OPTIONS_OPERATOR
    before = YADN('"spam"', ((m.Token.QUALIFIER, 'spam'),))
    alloweds = [
        m.Token.QUALIFIER,
        m.Token.QUALIFIER_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_pool_degen_operator():
    """Given a pool degeneration operator, return the correct tokens for
    the YADN.
    """
    token = m.Token.POOL_DEGEN_OPERATOR
    before = YADN('[3,3]', ((m.Token.POOL, (3, 3)),))
    alloweds = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


def test_pool_gen_operator():
    """Given a pool generation operator, return the correct tokens for
    the YADN.
    """
    token = m.Token.POOL_GEN_OPERATOR
    before = YADN('3', ((m.Token.NUMBER, 3),))
    alloweds = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


def test_pool_operator():
    """Given a pool operator, return the correct tokens for
    the YADN.
    """
    token = m.Token.POOL_OPERATOR
    before = YADN('[3,3]', ((m.Token.POOL, (3, 3)),))
    alloweds = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


def test_pool():
    """Given a pool, return the correct tokens for the YADN."""
    token = YADN('[3,3]', ((m.Token.POOL, (3, 3)),))
    before = YADN('', ())
    alloweds = [
        m.Token.GROUP_CLOSE,
        m.Token.POOL_OPERATOR,
        m.Token.POOL_DEGEN_OPERATOR,
        m.Token.ROLL_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_qualifier():
    """Given a qualifier, return the correct tokens for the YADN."""
    token = YADN('"spam"', ((m.Token.QUALIFIER, 'spam'),))
    before = YADN('', ())
    alloweds = [
        m.Token.OPTIONS_OPERATOR,
        m.Token.ROLL_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_roll_delimiter():
    """Given a roll delimiter, return the correct tokens for the YADN."""
    token = m.Token.ROLL_DELIMITER
    before = YADN('3', ((m.Token.NUMBER, 3),))
    alloweds = [
        m.Token.BOOLEAN,
        m.Token.GROUP_OPEN,
        m.Token.MAP_OPEN,
        m.Token.MAP,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.POOL,
        m.Token.POOL_OPEN,
        m.Token.QUALIFIER,
        m.Token.QUALIFIER_DELIMITER,
        m.Token.U_POOL_DEGEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


def test_u_pool_degen_operator():
    """Given an unary pool degeneration operator, return the correct
    tokens for the YADN.
    """
    token = m.Token.U_POOL_DEGEN_OPERATOR
    before = YADN('', ())
    alloweds = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.POOL,
        m.Token.POOL_OPEN,
        m.Token.U_POOL_DEGEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)


# Roll test case.
def test_start_roll():
    """Given a roll delimiter, return the correct tokens for the YADN."""
    token = YADN('', ())
    before = YADN('', ())
    alloweds = [
        m.Token.BOOLEAN,
        m.Token.GROUP_OPEN,
        m.Token.MAP_OPEN,
        m.Token.MAP,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.POOL,
        m.Token.POOL_OPEN,
        m.Token.QUALIFIER,
        m.Token.QUALIFIER_DELIMITER,
        m.Token.U_POOL_DEGEN_OPERATOR,
    ]
    lexer_test(token, before, alloweds)



"""
test_parse
~~~~~~~~~~

Unit tests for the yadr.parse module.
"""
from yadr import operator as yo
from yadr import parser as p
from yadr.model import Token


# Core test functions.
def parser_test(expected, tokens):
    parser = p.Parser()
    assert parser.parse(tokens) == expected


# Identity test cases.
def test_boolean():
    """Parse a boolean."""
    exp = True
    tokens = (
        (Token.BOOLEAN, True),
    )
    parser_test(exp, tokens)


def test_number():
    "Parse a number."
    exp = 5
    tokens = (
        (Token.NUMBER, 5),
    )
    parser_test(exp, tokens)


def test_pool():
    "Parse a pool."
    exp = (3, 3, 4)
    tokens = (
        (Token.POOL, (3, 3, 4)),
    )
    parser_test(exp, tokens)


def test_quantifier():
    """Parse a qualifier."""
    exp = 'spam'
    tokens = (
        (Token.QUALIFIER, 'spam'),
    )
    parser_test(exp, tokens)


# Basic arithmetic tests.
def test_addition():
    """Parse basic addition."""
    exp = 5
    tokens = (
        (Token.NUMBER, 3),
        (Token.AS_OPERATOR, '+'),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_division():
    "Perform basic division."
    exp = 2
    tokens = (
        (Token.NUMBER, 4),
        (Token.MD_OPERATOR, '/'),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_division_rounds_down():
    "Division rounds down to the nearest integer (floor division)."
    exp = 2
    tokens = (
        (Token.NUMBER, 5),
        (Token.MD_OPERATOR, '/'),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_exponentiation():
    "Perform basic division."
    exp = 9
    tokens = (
        (Token.NUMBER, 3),
        (Token.EX_OPERATOR, '^'),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_modulo():
    "Perform basic modulo."
    exp = 1
    tokens = (
        (Token.NUMBER, 3),
        (Token.MD_OPERATOR, '%'),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_multiplication():
    "Perform basic multiplication."
    exp = 6
    tokens = (
        (Token.NUMBER, 3),
        (Token.MD_OPERATOR, '*'),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_subtraction():
    "Perform basic subtraction."
    exp = 1
    tokens = (
        (Token.NUMBER, 3),
        (Token.AS_OPERATOR, '-'),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


# Choice and related operator tests.
def test_choice():
    """Perform choice operation."""
    exp = 'eggs'
    tokens = (
        (Token.BOOLEAN, False),
        (Token.CHOICE_OPERATOR, '?'),
        (Token.QUALIFIER, 'spam'),
        (Token.OPTIONS_OPERATOR, ':'),
        (Token.QUALIFIER, 'eggs'),
    )
    parser_test(exp, tokens)


def test_greater_than():
    """Perform greater than comparison."""
    exp = True
    tokens = (
        (Token.NUMBER, 3),
        (Token.COMPARISON_OPERATOR, '>'),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_greater_than_or_equal():
    """Perform greater than or equal comparison."""
    exp = True
    tokens = (
        (Token.NUMBER, 3),
        (Token.COMPARISON_OPERATOR, '>='),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_less_than():
    """Perform less than comparison."""
    exp = False
    tokens = (
        (Token.NUMBER, 3),
        (Token.COMPARISON_OPERATOR, '<'),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_less_than_or_equal():
    """Perform less than comparison."""
    exp = False
    tokens = (
        (Token.NUMBER, 3),
        (Token.COMPARISON_OPERATOR, '<='),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_equal():
    """Perform less than comparison."""
    exp = False
    tokens = (
        (Token.NUMBER, 3),
        (Token.COMPARISON_OPERATOR, '=='),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_not_equal():
    """Perform less than comparison."""
    exp = True
    tokens = (
        (Token.NUMBER, 3),
        (Token.COMPARISON_OPERATOR, '!='),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_options_operator():
    """Create a choice options."""
    exp = ('spam', 'eggs')
    tokens = (
        (Token.QUALIFIER, 'spam'),
        (Token.OPTIONS_OPERATOR, ':'),
        (Token.QUALIFIER, 'eggs'),
    )
    parser_test(exp, tokens)


# Dice operator tests.
def test_concat(mocker):
    """Concatenate dice rolls."""
    mocker.patch('random.randint', side_effect=[1, 11, 3])
    exp = 113
    tokens = (
        (Token.NUMBER, 3),
        (Token.DICE_OPERATOR, 'dc'),
        (Token.NUMBER, 12),
    )
    parser_test(exp, tokens)


def test_die(mocker):
    """Roll a die."""
    mocker.patch('random.randint', side_effect=[1, 2, 2])
    exp = 5
    tokens = (
        (Token.NUMBER, 3),
        (Token.DICE_OPERATOR, 'd'),
        (Token.NUMBER, 6),
    )
    parser_test(exp, tokens)


def test_exploding_die(mocker):
    """Roll an exploding die."""
    mocker.patch('random.randint', side_effect=[1, 2, 4, 2, 2])
    exp = 11
    tokens = (
        (Token.NUMBER, 4),
        (Token.DICE_OPERATOR, 'd!'),
        (Token.NUMBER, 4),
    )
    parser_test(exp, tokens)


def test_keep_high_die(mocker):
    """Roll dice and keep the highest."""
    mocker.patch('random.randint', side_effect=[1, 5, 3])
    exp = 5
    tokens = (
        (Token.NUMBER, 3),
        (Token.DICE_OPERATOR, 'dh'),
        (Token.NUMBER, 6),
    )
    parser_test(exp, tokens)


def test_keep_low_die(mocker):
    """Roll dice and keep the lowest."""
    mocker.patch('random.randint', side_effect=[1, 5, 3])
    exp = 1
    tokens = (
        (Token.NUMBER, 3),
        (Token.DICE_OPERATOR, 'dl'),
        (Token.NUMBER, 6),
    )
    parser_test(exp, tokens)


def test_wild_die(mocker):
    """Roll dice and keep the lowest."""
    mocker.patch('random.randint', side_effect=[4, 2, 1, 6])
    exp = 13
    tokens = (
        (Token.NUMBER, 4),
        (Token.DICE_OPERATOR, 'dw'),
        (Token.NUMBER, 6),
    )
    parser_test(exp, tokens)


# Pool generation operator tests.
def test_dice_pool(mocker):
    """Roll dice and keep the highest."""
    mocker.patch('random.randint', side_effect=[1, 5, 3])
    exp = (1, 5, 3)
    tokens = (
        (Token.NUMBER, 3),
        (Token.POOL_GEN_OPERATOR, 'g'),
        (Token.NUMBER, 6),
    )
    parser_test(exp, tokens)


def test_exploding_pool(mocker):
    """Roll an exploding dice pool."""
    mocker.patch('random.randint', side_effect=[1, 5, 3])
    exp = (1, 5, 3)
    tokens = (
        (Token.NUMBER, 3),
        (Token.POOL_GEN_OPERATOR, 'g!'),
        (Token.NUMBER, 6),
    )
    parser_test(exp, tokens)


# Unary pool degeneration operator tests.
def test_pool_concatenate(mocker):
    """Concatenate the dice in a pool."""
    mocker.patch('random.randint', side_effect=[1, 5, 3])
    exp = 153
    tokens = (
        (Token.U_POOL_DEGEN_OPERATOR, 'C'),
        (Token.NUMBER, (1, 5, 3)),
    )
    parser_test(exp, tokens)


def test_pool_count(mocker):
    """Count the dice in a pool."""
    mocker.patch('random.randint', side_effect=[1, 5, 3])
    exp = 3
    tokens = (
        (Token.U_POOL_DEGEN_OPERATOR, 'N'),
        (Token.NUMBER, (1, 5, 3)),
    )
    parser_test(exp, tokens)


def test_pool_sum(mocker):
    """Count the dice in a pool."""
    mocker.patch('random.randint', side_effect=[1, 5, 3])
    exp = 9
    tokens = (
        (Token.U_POOL_DEGEN_OPERATOR, 'S'),
        (Token.NUMBER, (1, 5, 3)),
    )
    parser_test(exp, tokens)


# Pool degeneration operators.
def test_count_successes():
    """Keep a values that aren't a given number in the pool."""
    exp = 2
    tokens = (
        (Token.POOL, (5, 8, 1, 9)),
        (Token.POOL_DEGEN_OPERATOR, 'ns'),
        (Token.NUMBER, 8),
    )
    parser_test(exp, tokens)


def test_count_successes_with_botch():
    """Keep a values that aren't a given number in the pool."""
    exp = 1
    tokens = (
        (Token.POOL, (5, 8, 1, 9)),
        (Token.POOL_DEGEN_OPERATOR, 'nb'),
        (Token.NUMBER, 8),
    )
    parser_test(exp, tokens)


# Pool operators.
def test_pool_keep_above():
    """Keep a values above a given number in the pool."""
    exp = (8, 9)
    tokens = (
        (Token.POOL, (5, 8, 1, 9)),
        (Token.POOL_OPERATOR, 'pa'),
        (Token.NUMBER, 6),
    )
    parser_test(exp, tokens)


def test_pool_keep_below():
    """Keep a values below a given number in the pool."""
    exp = (5, 1)
    tokens = (
        (Token.POOL, (5, 8, 1, 9)),
        (Token.POOL_OPERATOR, 'pb'),
        (Token.NUMBER, 6),
    )
    parser_test(exp, tokens)


def test_pool_cap():
    """Cap the values in a pool at a given value."""
    exp = (5, 7, 1, 7)
    tokens = (
        (Token.POOL, (5, 8, 1, 9)),
        (Token.POOL_OPERATOR, 'pc'),
        (Token.NUMBER, 7),
    )
    parser_test(exp, tokens)


def test_pool_floor():
    """Floor the values in a pool at a given value."""
    exp = (8, 8, 8, 9)
    tokens = (
        (Token.POOL, (5, 8, 1, 9)),
        (Token.POOL_OPERATOR, 'pf'),
        (Token.NUMBER, 8),
    )
    parser_test(exp, tokens)


def test_pool_keep_high():
    """Keep a number of highest values from the pool."""
    exp = (8, 9)
    tokens = (
        (Token.POOL, (5, 8, 1, 9)),
        (Token.POOL_OPERATOR, 'ph'),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_pool_keep_low():
    """Keep a number of highest values from the pool."""
    exp = (5, 1)
    tokens = (
        (Token.POOL, (5, 8, 1, 9)),
        (Token.POOL_OPERATOR, 'pl'),
        (Token.NUMBER, 2),
    )
    parser_test(exp, tokens)


def test_pool_modulo():
    """Perform a modulo on each member."""
    exp = (0, 3, 1, 4)
    tokens = (
        (Token.POOL, (5, 8, 1, 9)),
        (Token.POOL_OPERATOR, 'p%'),
        (Token.NUMBER, 5),
    )
    parser_test(exp, tokens)


def test_pool_remove():
    """Keep a values that aren't a given number in the pool."""
    exp = (5, 1, 9)
    tokens = (
        (Token.POOL, (5, 8, 1, 9)),
        (Token.POOL_OPERATOR, 'pr'),
        (Token.NUMBER, 8),
    )
    parser_test(exp, tokens)


# Dice mapping tests.
def test_map():
    """Parse and store a dice map."""
    # Expected value.
    exp = 'success'

    # Test data and state.
    parser = p.Parser()
    tokens = (
        (Token.MAP, ('spam', {1: 'none', 2: 'success'})),
    )
    _ = parser.parse(tokens)

    # Run test.
    assert parser.dice_map['spam'][2] == exp


# Mapping operator tests.
def test_map_result(mocker):
    """Given a map and a roll with a dice map operator, return the
    correct result.
    """
    # Expected value.
    exp = 'success'

    # Test data and state.
    parser = p.Parser()
    mocker.patch('random.randint', side_effect=(3, ))
    parser.dice_map['_test_map_result'] = {
        1: 'none',
        2: 'success',
        3: 'success',
        4: 'success success',
    }
    tokens = (
        (Token.NUMBER, 1),
        (Token.DICE_OPERATOR, 'd'),
        (Token.NUMBER, 4),
        (Token.MAPPING_OPERATOR, 'm'),
        (Token.QUALIFIER, '_test_map_result'),
    )

    # Run test.
    assert parser.parse(tokens) == exp


def test_map_result_pool(mocker):
    """Given a map and a roll with a dice map operator, return the
    correct result.
    """
    # Expected value.
    exp = ('success', 'none')

    # Test data and state.
    parser = p.Parser()
    mocker.patch('random.randint', side_effect=(3, 1, ))
    parser.dice_map['_test_map_result'] = {
        1: 'none',
        2: 'success',
        3: 'success',
        4: 'success success',
    }
    tokens = (
        (Token.NUMBER, 2),
        (Token.DICE_OPERATOR, 'g'),
        (Token.NUMBER, 4),
        (Token.MAPPING_OPERATOR, 'm'),
        (Token.QUALIFIER, '_test_map_result'),
    )

    # Run test.
    assert parser.parse(tokens) == exp


# Test rolls and result tests.
def test_roll_delimiter(mocker):
    """Return the result of two rolls."""
    mocker.patch('random.randint', side_effect=(1, 1, 3, 8))
    exp = ((1, 1, 3), 10)
    tokens = (
        (Token.NUMBER, 3),
        (Token.POOL_GEN_OPERATOR, 'g'),
        (Token.NUMBER, 8),
        (Token.ROLL_DELIMITER, ';'),
        (Token.NUMBER, 1),
        (Token.DICE_OPERATOR, 'd'),
        (Token.NUMBER, 10),
        (Token.AS_OPERATOR, '+'),
        (Token.NUMBER, 2)
    )
    parser_test(exp, tokens)


def test_roll_delimiter_remove_none_result(mocker):
    """Return the result of two rolls."""
    mocker.patch('random.randint', side_effect=(2,))
    exp = 'lose'
    tokens = (
        (Token.MAP, (
            'spam',
            {1: "win", 2: "lose"}
        )),
        (Token.ROLL_DELIMITER, ';'),
        (Token.NUMBER, 1),
        (Token.DICE_OPERATOR, 'd'),
        (Token.NUMBER, 2),
        (Token.MAPPING_OPERATOR, 'm'),
        (Token.QUALIFIER, 'spam')
    )
    parser_test(exp, tokens)


# Test order of precedence.
def test_can_perform_multiple_operations():
    """The parser can parse statements with multiple operators."""
    exp = 9
    tokens = (
        (Token.NUMBER, 3),
        (Token.AS_OPERATOR, '+'),
        (Token.NUMBER, 2),
        (Token.AS_OPERATOR, '+'),
        (Token.NUMBER, 4),
    )
    parser_test(exp, tokens)


def test_exponentiation_before_multiplication():
    """Multiplication should occur before addition."""
    exp = 48
    tokens = (
        (Token.NUMBER, 3),
        (Token.MD_OPERATOR, '*'),
        (Token.NUMBER, 2),
        (Token.EX_OPERATOR, '^'),
        (Token.NUMBER, 4),
    )
    parser_test(exp, tokens)


def test_multiplication_before_addition():
    """Multiplication should occur before addition."""
    exp = 11
    tokens = (
        (Token.NUMBER, 3),
        (Token.AS_OPERATOR, '+'),
        (Token.NUMBER, 2),
        (Token.MD_OPERATOR, '*'),
        (Token.NUMBER, 4),
    )
    parser_test(exp, tokens)


def test_parens_before_multiplication():
    """Parentheses should occur before multiplication."""
    exp = 18
    tokens = (
        (Token.NUMBER, 3),
        (Token.MD_OPERATOR, '*'),
        (Token.GROUP_OPEN, '('),
        (Token.NUMBER, 2),
        (Token.AS_OPERATOR, '+'),
        (Token.NUMBER, 4),
        (Token.GROUP_CLOSE, ')'),
    )
    parser_test(exp, tokens)


def test_pool_generation_happens_before_pool_degeneration(mocker):
    """Pool generation should happen before pool degeneration
    in order of operations.
    """
    # Expected value.
    exp = 20

    # Test data and state.
    rand_results = [5, 4, 6, 2, 3]
    mocker.patch('random.randint', side_effect=rand_results)
    tokens = (
        (Token.U_POOL_DEGEN_OPERATOR, 'S'),
        (Token.NUMBER, 5),
        (Token.POOL_GEN_OPERATOR, 'g'),
        (Token.NUMBER, 6),
    )

    # Run test.
    parser_test(exp, tokens)


def test_unary_pool_degeneration_happens_before_operators(mocker):
    """Unary pool degeneration should happen before operators
    in order of operations.
    """
    # Expected value.
    exp = 25

    # Test data and state.
    rand_results = [5, 4, 6, 2, 3]
    mocker.patch('random.randint', side_effect=rand_results)
    tokens = (
        (Token.U_POOL_DEGEN_OPERATOR, 'S'),
        (Token.NUMBER, 5),
        (Token.POOL_GEN_OPERATOR, 'g'),
        (Token.NUMBER, 6),
        (Token.AS_OPERATOR, '+'),
        (Token.NUMBER, 5),
    )

    # Run test.
    parser_test(exp, tokens)


def test_pool_operation_happens_before_pool_degeneration():
    """Pool operations should happen before pool degeneration
    in order of operations.
    """
    # Expected value.
    exp = 1

    # Test data and state.
    tokens = (
        (Token.POOL, (5, 4, 6, 2, 3)),
        (Token.POOL_OPERATOR, 'pb'),
        (Token.NUMBER, 5),
        (Token.POOL_DEGEN_OPERATOR, 'ns'),
        (Token.NUMBER, 5),
    )

    # Run test.
    parser_test(exp, tokens)



"""
test_encoder
~~~~~~~~~~~~

Unit tests for the yadr.encoder module.
"""
from yadr import encode as e
from yadr import model as m


# Test Encoder.encode().
def test_boolean():
    """An bool becomes a string containing the YADN boolean."""
    exp = 'T'
    data = True
    encode_test(exp, data)


def test_compound_result():
    """A CompoundResult becomes a string of roll delimited values."""
    exp = '[1, 1, 3, 8]; 3'
    data = m.CompoundResult((
        (1, 1, 3, 8),
        3
    ))
    encode_test(exp, data)


def test_number():
    """An int becomes a string containing the YADN number."""
    exp = '3'
    data = 3
    encode_test(exp, data)


def test_pool():
    """A tuple of integers becomes a string containing the
    YADN pool.
    """
    exp = '[1, 1, 3, 8]'
    data = (1, 1, 3, 8)
    encode_test(exp, data)


def test_qualifier():
    """A string becomes a double-quoted string."""
    exp = '"spam"'
    data = 'spam'
    encode_test(exp, data)


def encode_test(expected, data):
    """Run a test on :meth:`Encoder.encode`."""
    encoder = e.Encoder()
    assert encoder.encode(data) == expected



"""
test_maps
~~~~~~~~~

Unittests for the yadr.maps module.
"""
from collections import namedtuple
from functools import partial

import pytest

from yadr import maps
from yadr import model as m
from yadr.model import Token


# Registration.
cases = {}


class reg:
    def __init__(self, key):
        self.key = key

    def __call__(self, fn):
        new = partial(fn, token=self.key)
        cases[self.key] = new
        return fn


# Common test case utilities.
YADN = namedtuple('YADN', 'yadn tokens')


def get_yadn(token):
    """Build a :class:`YADN` object for each valid symbol for the token."""
    # If a YADN object was passed, just yield it back.
    if isinstance(token, YADN):
        yield token

    # Otherwise get the symbols for the token type, build and then
    # yield the YADN object for each of the symbols.
    else:
        symbols = m.yadn_symbols_raw[token].split()
        for symbol in symbols:
            yadn = symbol
            value = symbol
            if token == m.Token.NUMBER:
                value = int(value)
            elif token == m.Token.BOOLEAN:
                value = True
                if symbol == 'F':
                    value = False
            yield YADN(yadn, ((token, value),))


# Build the test data for testing each type of token.
@reg(m.Token.KV_DELIMITER)
def build_kv_delimiter_cases(token):
    yadns = get_yadn(token)
    for yadn in yadns:
        yield YADN(f'{yadn.yadn}"spam"}}', (
            *yadn.tokens,
            (m.Token.QUALIFIER, 'spam'),
            (m.Token.MAP_CLOSE, '}'),
        ))


@reg(m.Token.NEGATIVE_SIGN)
def build_negative_sign_cases(token):
    yield YADN('-2', ((m.Token.NUMBER, -2),))


@reg(m.Token.NAME_DELIMITER)
def build_name_delimiter_cases(token):
    yadns = get_yadn(token)
    for yadn in yadns:
        yield YADN(f'{yadn.yadn}2:"spam"}}', (
            *yadn.tokens,
            (m.Token.NUMBER, 2),
            (m.Token.KV_DELIMITER, ':'),
            (m.Token.QUALIFIER, 'spam'),
            (m.Token.MAP_CLOSE, '}'),
        ))


@reg(m.Token.PAIR_DELIMITER)
def build_pair_delimiter_cases(token):
    yadns = get_yadn(token)
    for yadn in yadns:
        yield YADN(f'{yadn.yadn}2:"spam"}}', (
            *yadn.tokens,
            (m.Token.NUMBER, 2),
            (m.Token.KV_DELIMITER, ':'),
            (m.Token.QUALIFIER, 'spam'),
            (m.Token.MAP_CLOSE, '}'),
        ))


@reg(m.Token.QUALIFIER)
@reg(m.Token.QUALIFIER_DELIMITER)
def build_qualifier_cases(token=None):
    yield YADN('"spam"', ((m.Token.QUALIFIER, 'spam'),))


@reg(m.Token.MAP_CLOSE)
@reg(m.Token.NUMBER)
def build_token_only_case(token):
    for yadn in get_yadn(token):
        yield yadn


# Core test code.
def allowed_test(symbol, before, after):
    """Test a token that is allowed to follow the given symbol."""
    expected = (
        *before.tokens,
        *symbol.tokens,
        *after.tokens,
    )

    lexer = maps.Lexer()
    yadns = [
        f'{before.yadn}{symbol.yadn}{after.yadn}',
        f'{before.yadn}{symbol.yadn} {after.yadn}',
    ]

    for yadn in yadns:
        actual = lexer.lex(yadn)
        try:
            assert actual == expected
        except AssertionError as ex:
            cls = ex.__class__
            msg = f'{ex}\n{yadn}'
            raise cls(msg)


def disallowed_test(symbol, before, after):
    """Test a token that is not allowed to follow the given symbol"""
    expected = ValueError

    lexer = maps.Lexer()
    yadns = [
        f'{before.yadn}{symbol.yadn}{after.yadn}',
        f'{before.yadn}{symbol.yadn} {after.yadn}',
    ]

    for yadn in yadns:
        try:
            pytest.raises(expected, lexer.lex, yadn)
        except BaseException as ex:
            cls = ex.__class__
            msg = f'{ex}\n{yadn}'
            raise cls(msg)


def lexer_test(token, before, alloweds):
    """Common test of :meth:`Lexer.lex` for a token."""
    # Build list of disallowed tokens.
    disalloweds = [token for token in cases if token not in alloweds]

    # Since AS_OPERATORS contain -, the NEGATIVE_SIGN will be allowed
    # if AS_OPERATORS are allowed.
    if m.Token.AS_OPERATOR in alloweds:
        disalloweds = [
            token for token in disalloweds
            if token != m.Token.NEGATIVE_SIGN
        ]

    # Since AS_OPERATORS contain -, that AS_OPERATOR is allowed when
    # NEGATIVE_SIGN is allowed. This change does prevent us from testing
    # whether + is allowed, though.
    if m.Token.NEGATIVE_SIGN in alloweds:
        disalloweds = [
            token for token in disalloweds
            if token != m.Token.AS_OPERATOR
        ]

    # NUMBERS next to NUMBERS aren't separate tokens.
    if (
        token == m.Token.NUMBER
        or (
            isinstance(token, YADN)
            and token.tokens
            and token.tokens[0][0] == m.Token.NUMBER
        )
    ):
        disalloweds = [
            token for token in disalloweds
            if token != m.Token.NUMBER
        ]

    # Test each of the possible symbols for the token.
    symbols = get_yadn(token)
    for symbol in symbols:

        # Test alloweds.
        for key in alloweds:
            afters = cases[key]()
            for after in afters:
                allowed_test(symbol, before, after)

        # Test disalloweds.
        for key in disalloweds:
            afters = cases[key]()
            for after in afters:
                disallowed_test(symbol, before, after)


# Symbol unit tests.
def test_kv_delimiter():
    """Given a key-value delimiter, return the proper tokens."""
    token = Token.KV_DELIMITER
    before = YADN(
        '{"spam"=1',
        (
            (m.Token.MAP_OPEN, '{'),
            (m.Token.QUALIFIER, 'spam'),
            (m.Token.NAME_DELIMITER, '='),
            (m.Token.NUMBER, 1),
        )
    )
    alloweds = [
        Token.NEGATIVE_SIGN,
        Token.NUMBER,
        Token.QUALIFIER,
        Token.QUALIFIER_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_map_close():
    """Given a map close character, return the proper tokens."""
    token = Token.MAP_CLOSE
    before = YADN(
        '{"spam"=1:"eggs"',
        (
            (m.Token.MAP_OPEN, '{'),
            (m.Token.QUALIFIER, 'spam'),
            (m.Token.NAME_DELIMITER, '='),
            (m.Token.NUMBER, 1),
            (m.Token.KV_DELIMITER, ':'),
            (m.Token.QUALIFIER, 'eggs'),
        )
    )
    alloweds = []
    lexer_test(token, before, alloweds)


def test_map_open():
    """Given a map open character, return the proper tokens."""
    token = Token.MAP_OPEN
    before = YADN('', ())
    alloweds = [
        Token.MAP_CLOSE,
        Token.QUALIFIER,
        Token.QUALIFIER_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_name_delimiter():
    """Given a name delimiter character, return the proper tokens."""
    token = Token.NAME_DELIMITER
    before = YADN(
        '{"spam"',
        (
            (m.Token.MAP_OPEN, '{'),
            (m.Token.QUALIFIER, 'spam'),
        )
    )
    alloweds = [
        Token.NEGATIVE_SIGN,
        Token.NUMBER,
    ]
    lexer_test(token, before, alloweds)


def test_number():
    """Given a number, return the proper tokens."""
    token = m.Token.NUMBER
    before = YADN(
        '{"spam"=',
        (
            (m.Token.MAP_OPEN, '{'),
            (m.Token.QUALIFIER, 'spam'),
            (m.Token.NAME_DELIMITER, '='),
        )
    )
    alloweds = [
        Token.KV_DELIMITER,
        Token.MAP_CLOSE,
        Token.PAIR_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_number_with_negative_sign():
    """Given a negative sign character, return the proper tokens."""
    token = YADN('-2', ((m.Token.NUMBER, -2),))
    before = YADN(
        '{"spam"=',
        (
            (m.Token.MAP_OPEN, '{'),
            (m.Token.QUALIFIER, 'spam'),
            (m.Token.NAME_DELIMITER, '='),
        )
    )
    alloweds = [
        Token.KV_DELIMITER,
        Token.MAP_CLOSE,
        Token.PAIR_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


def test_pair_delimiter():
    """Given a pair delimiter, return the proper tokens."""
    token = m.Token.PAIR_DELIMITER
    before = YADN(
        '{"spam"=1:"eggs"',
        (
            (m.Token.MAP_OPEN, '{'),
            (m.Token.QUALIFIER, 'spam'),
            (m.Token.NAME_DELIMITER, '='),
            (m.Token.NUMBER, 1),
            (m.Token.KV_DELIMITER, ':'),
            (m.Token.QUALIFIER, 'eggs'),
        )
    )
    alloweds = [
        Token.NEGATIVE_SIGN,
        Token.NUMBER,
    ]
    lexer_test(token, before, alloweds)


def test_qualifier():
    """Given a qualifier, return the proper tokens."""
    token = m.Token.QUALIFIER
    before = YADN('{', ((m.Token.MAP_OPEN, '{'),))
    alloweds = [
        Token.MAP_CLOSE,
        Token.NAME_DELIMITER,
        Token.PAIR_DELIMITER,
    ]
    lexer_test(token, before, alloweds)


# Parsing test cases.
def test_parser():
    """A basic dice mapping can be parsed."""
    tokens = (
        (Token.MAP_OPEN, '{'),
        (Token.QUALIFIER, 'name'),
        (Token.NAME_DELIMITER, '='),
        (Token.NUMBER, 1),
        (Token.KV_DELIMITER, ':'),
        (Token.QUALIFIER, 'none'),
        (Token.PAIR_DELIMITER, ','),
        (Token.NUMBER, 2),
        (Token.KV_DELIMITER, ':'),
        (Token.QUALIFIER, 'success'),
        (Token.PAIR_DELIMITER, ','),
        (Token.NUMBER, 3),
        (Token.KV_DELIMITER, ':'),
        (Token.QUALIFIER, 'success'),
        (Token.PAIR_DELIMITER, ','),
        (Token.NUMBER, 4),
        (Token.KV_DELIMITER, ':'),
        (Token.QUALIFIER, 'success success'),
        (Token.PAIR_DELIMITER, ','),
        (Token.MAP_CLOSE, '}'),
    )
    parser = maps.Parser()
    assert parser.parse(tokens) == (
        'name',
        {
            1: "none",
            2: "success",
            3: "success",
            4: "success success",
        }
    )


def test_parser_with_numbers():
    """A basic dice mapping can be parsed."""
    tokens = (
        (Token.MAP_OPEN, '{'),
        (Token.QUALIFIER, 'name'),
        (Token.NAME_DELIMITER, '='),
        (Token.NUMBER, 1),
        (Token.KV_DELIMITER, ':'),
        (Token.NUMBER, -1),
        (Token.PAIR_DELIMITER, ','),
        (Token.NUMBER, 2),
        (Token.KV_DELIMITER, ':'),
        (Token.NUMBER, 0),
        (Token.PAIR_DELIMITER, ','),
        (Token.NUMBER, 3),
        (Token.KV_DELIMITER, ':'),
        (Token.NUMBER, 1),
        (Token.MAP_CLOSE, '}'),
    )
    parser = maps.Parser()
    assert parser.parse(tokens) == (
        'name',
        {
            1: -1,
            2: 0,
            3: 1,
        }
    )



"""
test_yadr
~~~~~~~~~

Unit tests for the yadr.yadr module.
"""
from yadr import __main__, yadr


# Test yadr.roll().
def test_roll(mocker):
    """Execute a YADN string."""
    exp = 11
    yadn = ('3d6',)
    dice = (4, 4, 3)
    roll_test(exp, yadn, dice, mocker)


def test_roll_with_default_dice_map(mocker):
    """Execute a YADN string using a default dice map."""
    exp = '-'
    yadn = ('1d3m"fate"',)
    dice = (1,)
    roll_test(exp, yadn, dice, mocker)


def test_roll_with_yadn_output(mocker):
    """Execute a YADN string returning a YADN string."""
    exp = '11'
    yadn = ('3d6', True)
    dice = (4, 4, 3)
    roll_test(exp, yadn, dice, mocker)


def roll_test(exp, params, dice, mocker):
    mocker.patch('random.randint', side_effect=dice)
    assert yadr.roll(*params) == exp


# Test parse_cli().
def test_parse_cli(mocker, capsys):
    """Execute YADN from the command line."""
    cmd = ['python -m yadr', '3d6']
    dice = (4, 4, 3)
    result = cli_test(cmd, dice, mocker, capsys)
    assert result == '11\n'


def test_parse_cli_with_dice_maps(mocker, capsys):
    """The -m option followed by a file path should load in the
    dice maps at the given location.
    """
    cmd = [
        'python -m yadr',
        '2g3m"fudge"',
        '-m',
        'tests/data/__test_dice_map.txt'
    ]
    dice = (3, 2)
    result = cli_test(cmd, dice, mocker, capsys)
    assert result == '["+", ""]\n'


def test_parse_cli_with_listing_default_maps(mocker, capsys):
    """The -l option will list the default dice maps."""
    cmd = ['python -m yadr', '-l']
    dice = ()
    result = cli_test(cmd, dice, mocker, capsys)

    # Build the expected value.
    default_map_loc = 'src/yadr/data/dice_maps.yadn'
    with open(default_map_loc) as fh:
        lines = fh.readlines()
    lines = [line for line in lines if '=' in line]
    lines = [line.split('"')[1] for line in lines]
    expected = '\n'.join(lines) + '\n'

    assert result == expected


def cli_test(cmd, dice, mocker, capsys):
    """Test the output of running `yadr` from the command line."""
    # Set up the test.
    mocker.patch('sys.argv', cmd)
    mocker.patch('random.randint', side_effect=dice)

    # Run the test.
    __main__.parse_cli()

    # Capture and return the test result.
    captured = capsys.readouterr()
    return captured.out



"""
operator
~~~~~~~~

Operators for handling the dice part of dice notation.
"""
import operator
import random
import secrets
from collections.abc import Callable, Sequence


# Result types for annotation.
Options = tuple[str, str]
Pool = Sequence[int]


# Operation types for annotation.
CompOp = Callable[[int, int], bool]
OptionsOp = Callable[[str, str], Options]
ChoiceOp = Callable[[bool, Options], str]
PoolGenOp = Callable[[int, int], Pool]
DiceOp = Callable[[int, int], int]
MathOp = Callable[[int, int], int]
PoolOp = Callable[[Pool, int], Pool]
PoolDegenOp = Callable[[Pool, int], int]
UPoolDegenOp = Callable[[Pool], int]
Operation = Callable


# Registration.
ops: dict[str, Operation] = {
    '^': operator.pow,
    '*': operator.mul,
    '/': operator.floordiv,
    '%': operator.mod,
    '+': operator.add,
    '-': operator.sub,
    '>': operator.gt,
    '<': operator.le,
    '>=': operator.ge,
    '<=': operator.le,
    '==': operator.eq,
    '!=': operator.ne,
}


class operation:
    """A registration decorator for operations.

    :param symbol: The string used to refer to the operation.
    :return: None.
    :rtype: NoneType
    """
    def __init__(self, symbol: str) -> None:
        self.symbol = symbol

    def __call__(self, fn: Operation) -> Operation:
        """Register the operation.

        :param fn: The decorated function. It's sent automatically
            when :class:`operator.operation` is used as a decorator.
        :return: The decorated :class:`collections.abc.Callable`.
        :rtype: Callable
        """
        ops[self.symbol] = fn
        return fn


# Choice operators.
@operation(':')
def choice_options(a: str, b: str) -> Options:
    """Create the options for a choice.

    :ref:`YADN` reference: :ref:`choice_options`

    :param a: The qualifier for the true condition of a choice.
    :param b: The qualifier for the false condition of a choice.
    :return: The qualifiers as a :class:`tuple`.
    :rtype: tuple

    Usage::

        >>> choice_options('success', 'failure')
        ('success', 'failure')
    """
    return (a, b)


@operation('?')
def choice(boolean: bool, options: Options) -> str:
    """Make a choice.

    :ref:`YADN` reference: :ref:`choice`

    :param boolean: The decision as a :class:`bool`.
    :param options: The two options to pick from.
    :return: The chosen option as a :class:`str`.
    :rtype: str

    Usage::

        >>> choice(False, ('spam', 'eggs'))
        'eggs'
    """
    result = options[0]
    if not boolean:
        result = options[1]
    return result


# Dice operators.
@operation('dc')
def concat(num: int, size: int) -> int:
    """Concatenate the least significant digits.

    :ref:`YADN` reference: :ref:`concat`

    :param num: The number of dice to roll.
    :param size: The highest number that can be rolled on a die.
    :return: The concatenated least significant digits as an
        :class:`int`.
    :rtype: int

    Usage::

        >>> # This line is to ensure predictability for testing.
        >>> # Do not use outside of test cases.
        >>> _seed('spam')
        >>>
        >>> # Roll 2d10 as percentile dice. It's not quite right,
        >>> # since 00 should be 100, but the physical dice have
        >>> # that problem, too.
        >>> concat(2, 10)
        21
    """
    base = 10
    pool = dice_pool(num, size)
    pool = pool_modulo(pool, base)
    return pool_concatenate(pool)


@operation('d')
def die(num: int, size: int) -> int:
    """Roll a number of same-sized dice and return the result.

    :ref:`YADN` reference: :ref:`die`

    :param num: The number of dice to roll.
    :param size: The highest number that can be rolled on a die.
    :return: The sum of the result of each die as an :class:`int`.
    :rtype: int

    Usage::

        >>> # This line is to ensure predictability for testing.
        >>> # Do not use outside of test cases.
        >>> _seed('spam')
        >>>
        >>> # Roll 3d6.
        >>> die(3, 6)
        5

    """
    pool = dice_pool(num, size)
    return sum(pool)


@operation('d!')
def exploding_die(num: int, size: int) -> int:
    """Roll a number of exploding same-sized dice.

    :ref:`YADN` reference: :ref:`explode`

    :param num: The number of dice to roll.
    :param size: The highest number that can be rolled on a die.
    :return: The sum of the result of each die as an :class:`int`.
    :rtype: int

    Usage::

        >>> # This line is to ensure predictability for testing.
        >>> # Do not use outside of test cases.
        >>> _seed('spam')
        >>>
        >>> # Roll 5d!6.
        >>> exploding_die(5, 6)
        15

    """
    return sum(exploding_pool(num, size))


@operation('dh')
def keep_high_die(num: int, size: int) -> int:
    """Roll a number of dice and keep the highest.

    :ref:`YADN` reference: :ref:`keep_high`

    :param num: The number of dice to roll.
    :param size: The highest number that can be rolled on a die.
    :return: The highest value as an :class:`int`.
    :rtype: int

    Usage::

        >>> # This line is to ensure predictability for testing.
        >>> # Do not use outside of test cases.
        >>> _seed('spam')
        >>>
        >>> # Roll 5dh6.
        >>> keep_high_die(5, 6)
        5

    """
    pool = dice_pool(num, size)
    return max(pool)


@operation('dl')
def keep_low_die(num: int, size: int) -> int:
    """Roll a number of dice and keep the lowest.

    :ref:`YADN` reference: :ref:`keep_low`

    :param num: The number of dice to roll.
    :param size: The highest number that can be rolled on a die.
    :return: The lowest value as an :class:`int`.
    :rtype: int

    Usage::

        >>> # This line is to ensure predictability for testing.
        >>> # Do not use outside of test cases.
        >>> _seed('spam')
        >>>
        >>> # Roll 5dh6.
        >>> keep_low_die(5, 6)
        1

    """
    pool = dice_pool(num, size)
    return min(pool)


@operation('dw')
def wild_die(num: int, size: int) -> int:
    """Roll a number of same-sized dice and return the result, with
    one of the dice being the wild die.

    :ref:`YADN` reference: :ref:`wild_die`

    :param num: The number of dice to roll.
    :param size: The highest number that can be rolled on a die.
    :return: The sum of the values as an :class:`int`.
    :rtype: int

    Usage::

        >>> # This line is to ensure predictability for testing.
        >>> # Do not use outside of test cases.
        >>> _seed('spam')
        >>>
        >>> # Roll 5dw6.
        >>> wild_die(5, 6)
        0

    """
    wild = exploding_pool(1, size)
    regular = dice_pool(num - 1, size)
    if wild[0] == 1:
        return 0
    return sum((sum(wild), sum(regular)))


# Pool operators.
@operation('pc')
def pool_cap(pool: Pool, cap: int) -> Pool:
    """Cap the maximum value in a pool.

    :ref:`YADN` reference: :ref:`pool_cap`

    :param pool: A sequence of die values.
    :param cap: The maximum value of a die roll.
    :return: The resulting values as a :class:`tuple`.
    :rtype: tuple

    Usage::

        >>> # Roll [4, 10, 3, 5, 1, 9]pc6.
        >>> pool_cap([4, 10, 3, 5, 1, 9], 6)
        (4, 6, 3, 5, 1, 6)

    """
    result = []
    for value in pool:
        if value > cap:
            value = cap
        result.append(value)
    return tuple(result)


@operation('pf')
def pool_floor(pool: Pool, floor: int) -> Pool:
    """Floor the minimum value in a pool.

    :ref:`YADN` reference: :ref:`pool_floor`

    :param pool: A sequence of die values.
    :param floor: The minimum value of a die roll.
    :return: The resulting values as a :class:`tuple`.
    :rtype: tuple

    Usage::

        >>> # Roll [4, 10, 3, 5, 1, 9]pf6.
        >>> pool_floor([4, 10, 3, 5, 1, 9], 6)
        (6, 10, 6, 6, 6, 9)

    """
    result = []
    for value in pool:
        if value < floor:
            value = floor
        result.append(value)
    return tuple(result)


@operation('pa')
def pool_keep_above(pool: Pool, floor: int) -> Pool:
    """Discard all values in a pool below a given value.

    :ref:`YADN` reference: :ref:`pool_keep_above`

    :param pool: A sequence of die values.
    :param floor: The minimum value to keep.
    :return: The resulting values as a :class:`tuple`.
    :rtype: tuple

    Usage::

        >>> # Roll [4, 10, 3, 5, 1, 9]pa6.
        >>> pool_keep_above([4, 10, 3, 5, 1, 9], 6)
        (10, 9)

    """
    return tuple(n for n in pool if n >= floor)


@operation('pb')
def pool_keep_below(pool: Pool, ceiling: int) -> Pool:
    """Discard all values in a pool above a given value.

    :ref:`YADN` reference: :ref:`pool_keep_below`

    :param pool: A sequence of die values.
    :param ceiling: The maximum value to keep.
    :return: The resulting values as a :class:`tuple`.
    :rtype: tuple

    Usage::

        >>> # Roll [4, 10, 3, 5, 1, 9]pb6.
        >>> pool_keep_below([4, 10, 3, 5, 1, 9], 6)
        (4, 3, 5, 1)

    """
    return tuple(n for n in pool if n <= ceiling)


@operation('ph')
def pool_keep_high(pool: Pool, keep: int) -> Pool:
    """Keep a number of the highest dice.

    :ref:`YADN` reference: :ref:`pool_keep_high`

    :param pool: A sequence of die values.
    :param keep: The maximum value to keep.
    :return: The resulting values as a :class:`tuple`.
    :rtype: tuple

    Usage::

        >>> # Roll [4, 10, 3, 5, 1, 9]ph3.
        >>> pool_keep_high([4, 10, 3, 5, 1, 9], 3)
        (10, 5, 9)

    """
    pool = list(pool)
    remove = len(pool) - keep
    for _ in range(remove):
        low_value = max(pool)
        low_index = 0
        for i, n in enumerate(pool):
            if n < low_value:
                low_value = n
                low_index = i
        pool.pop(low_index)
    return tuple(pool)


@operation('pl')
def pool_keep_low(pool: Pool, keep: int) -> Pool:
    """Keep a number of the lowest dice.

    :ref:`YADN` reference: :ref:`pool_keep_low`

    :param pool: A sequence of die values.
    :param keep: The maximum value to keep.
    :return: The resulting values as a :class:`tuple`.
    :rtype: tuple

    Usage::

        >>> # Roll [4, 10, 3, 5, 1, 9]pl3.
        >>> pool_keep_low([4, 10, 3, 5, 1, 9], 3)
        (4, 3, 1)

    """
    pool = list(pool)
    remove = len(pool) - keep
    for _ in range(remove):
        high_value = min(pool)
        high_index = 0
        for i, n in enumerate(pool):
            if n > high_value:
                high_value = n
                high_index = i
        pool.pop(high_index)
    return tuple(pool)


@operation('p%')
def pool_modulo(pool: Pool, divisor: int) -> Pool:
    """Perform a modulo operation of each member.

    :ref:`YADN` reference: :ref:`pool_mod`

    :param pool: A sequence of die values.
    :param divisor: The maximum value to keep.
    :return: The resulting values as a :class:`tuple`.
    :rtype: tuple

    Usage::

        >>> # Roll [4, 10, 3, 5, 1, 9]p%3.
        >>> pool_modulo([4, 10, 3, 5, 1, 9], 3)
        (1, 1, 0, 2, 1, 0)

    """
    return tuple(n % divisor for n in pool)


@operation('pr')
def pool_remove(pool: Pool, cut: int) -> Pool:
    """Remove members of a pool of the given value.

    :ref:`YADN` reference: :ref:`pool_remove`

    :param pool: A sequence of die values.
    :param cut: The maximum value to keep.
    :return: The resulting values as a :class:`tuple`.
    :rtype: tuple

    Usage::

        >>> # Roll [4, 10, 3, 5, 1, 9]pr5.
        >>> pool_remove([4, 10, 3, 5, 1, 9], 5)
        (4, 10, 3, 1, 9)

    """
    return tuple(n for n in pool if n != cut)


# Pool degeneration operators.
@operation('C')
def pool_concatenate(pool: Pool) -> int:
    """Concatenate the dice in the pool.

    :ref:`YADN` reference: :ref:`pool_concat`

    :param pool: A sequence of die values.
    :return: The resulting value as an :class:`int`.
    :rtype: int

    Usage::

        >>> # Roll C[4, 10, 3, 5, 1, 9].
        >>> pool_concatenate([4, 10, 3, 5, 1, 9])
        4103519

    """
    str_value = ''.join((str(m) for m in pool))
    return int(str_value)


@operation('N')
def pool_count(pool: Pool) -> int:
    """Count the dice in the pool.

    :ref:`YADN` reference: :ref:`pool_count`

    :param pool: A sequence of die values.
    :return: The resulting value as an :class:`int`.
    :rtype: int

    Usage::

        >>> # Roll N[4, 10, 3, 5, 1, 9].
        >>> pool_count([4, 10, 3, 5, 1, 9])
        6

    """
    return len(pool)


@operation('S')
def pool_sum(pool: Pool) -> int:
    """Sum the dice in the pool.

    :ref:`YADN` reference: :ref:`pool_sum`

    :param pool: A sequence of die values.
    :return: The resulting value as an :class:`int`.
    :rtype: int

    Usage::

        >>> # Roll S[4, 10, 3, 5, 1, 9].
        >>> pool_sum([4, 10, 3, 5, 1, 9])
        32

    """
    return sum(pool)


@operation('ns')
def count_successes(pool: Pool, target: int) -> int:
    """Count the number of successes in the pool.

    :ref:`YADN` reference: :ref:`count_successes`

    :param pool: A sequence of die values.
    :param target: The target number for success.
    :return: The resulting value as an :class:`int`.
    :rtype: int

    Usage::

        >>> # Roll [4, 10, 3, 5, 1, 9]ns6.
        >>> count_successes([4, 10, 3, 5, 1, 9], 6)
        2

    """
    pool = pool_keep_above(pool, target)
    return len(pool)


@operation('nb')
def count_successes_with_botch(pool: Pool, target: int) -> int:
    """Count the number of successes in the pool. Then remove a success
    for each botch.

    :ref:`YADN` reference: :ref:`count_botch`

    :param pool: A sequence of die values.
    :param target: The target number for success.
    :return: The resulting value as an :class:`int`.
    :rtype: int

    Usage::

        >>> # Roll [4, 10, 3, 5, 1, 9]nb6.
        >>> count_successes_with_botch([4, 10, 3, 5, 1, 9], 6)
        1

    """
    botches = len([n for n in pool if n == 1])
    pool = pool_keep_above(pool, target)
    return len(pool) - botches


# Pool generation operator.
@operation('g')
def dice_pool(num: int, size: int) -> Pool:
    """Roll a dice pool.

    :ref:`YADN` reference: :ref:`dice_pool`

    :param num: The number of dice to roll.
    :param size: The highest number that can be rolled on a die.
    :return: The the values as a :class:`tuple`.
    :rtype: tuple

    Usage::

        >>> # This line is to ensure predictability for testing.
        >>> # Do not use outside of test cases.
        >>> _seed('spam')
        >>>
        >>> # Roll 5g6.
        >>> dice_pool(5, 6)
        (1, 1, 3, 5, 5)

    """
    return tuple(roll(size) for _ in range(num))


@operation('g!')
def exploding_pool(num: int, size: int) -> Pool:
    """Roll an exploding dice pool.

    :ref:`YADN` reference: :ref:`exploding_pool`

    :param num: The number of dice to roll.
    :param size: The highest number that can be rolled on a die.
    :return: The the values as a :class:`tuple`.
    :rtype: tuple

    Usage::

        >>> # This line is to ensure predictability for testing.
        >>> # Do not use outside of test cases.
        >>> _seed('spam')
        >>>
        >>> # Roll 5g!6.
        >>> exploding_pool(5, 6)
        (1, 1, 3, 5, 5)

    """
    return tuple(_explode(size) for n in range(num))


# Generation operators that use `secrets` instead of `random`.
def dice_pool_secrets(num: int, size: int) -> Pool:
    """Roll a dice pool using :mod:`secrets` instead of :mod:`random`.
    The extra security provided by :mod:`secrets` is unnecessary for
    the intended usage of :mod:`yadr`, but it's here if you want it.

    :ref:`YADN` reference: :ref:`dice_pool`

    :param num: The number of dice to roll.
    :param size: The highest number that can be rolled on a die.
    :return: The the values as a :class:`tuple`.
    :rtype: tuple

    .. warning::
        This function is experimental and may be removed from future
        versions of :mod:`yadr` without notice. Use at your own risk.
    """
    return tuple(_roll_secrets(size) for _ in range(num))


# Utility functions.
def _explode(size: int) -> int:
    """Explode the value of a die.

    :param size: The highest number that can be rolled on a die.
    :return: The the values as an :class:`int`.
    :rtype: int
    """
    result = roll(size)
    if result == size:
        result += _explode(size)
    return result


def _seed(seed: int | str | bytes) -> None:
    """Seed the random number generator for testing purposes.

    :param seed: A seed value for the random number generator.
    :return: None.
    :rtype: NoneType
    """
    if isinstance(seed, str):
        seed = bytes(seed, encoding='utf_8')
    if isinstance(seed, bytes):
        seed = int.from_bytes(seed, 'little')
    random.seed(seed)


# Die rolling.
def _roll_random(size: int) -> int:
    """Roll a die.

    :param size: The size of the die to roll.
    :returns: An :class:'int' object.
    :rtype: int
    """
    return random.randint(1, size)


def _roll_secrets(size: int) -> int:
    """Roll a die using :mod:`secrets`.

    :param size: The size of the die to roll.
    :returns: An :class:'int' object.
    :rtype: int
    """
    return secrets.randbelow(size) + 1


# This sets the function used by the generation operations to roll dice.
# If you want to use `secrets` instead of `random`, change this to
# :func:`_roll_secrets`. Since this is happening at the global level,
# you could, theoretically, run into thread safety issues when changing it,
# So be cautious.
roll = _roll_random



"""
test_operator
~~~~~~~~~~~~~

Dice operators for the `yadr` package.
"""
from yadr import operator as op


# Choice test cases.
def test_choice_options():
    """Generate choice options."""
    a = 'spam'
    b = 'eggs'
    assert op.choice_options(a, b) == (a, b,)


def test_choice():
    """Make a choice."""
    boolean = True
    options = ('spam', 'eggs')
    assert op.choice(boolean, options) == 'spam'


# Dice operation test cases.
def test_concat(mocker):
    """Concatenate the least significant digit of the dice."""
    mocker.patch('random.randint', side_effect=(3, 10, 4))
    num = 3
    size = 10
    assert op.concat(num, size) == 304


def test_die(mocker):
    """Given a number of dice and the size of the die, roll that many
    dice and return the result.
    """
    mocker.patch('random.randint', side_effect=(4, 2, 1))
    num = 2
    size = 6
    assert op.die(num, size) == 6


def test_exploding_die(mocker):
    """Given a number of dice and the size of the die, roll that many
    exploding dice and return the result.
    """
    mocker.patch('random.randint', side_effect=[2, 1, 4, 4, 3, 1, 4, 4, 2])
    num = 5
    size = 4
    assert op.exploding_die(num, size) == 25


def test_keep_high_die(mocker):
    """Given a number of dice and the size of the die, roll that many
    dice and keep the highest.
    """
    mocker.patch('random.randint', side_effect=[15, 3, 6, 18, 10, 20])
    num = 5
    size = 20
    assert op.keep_high_die(num, size) == 18


def test_keep_low_die(mocker):
    """Given a number of dice and the size of the die, roll that many
    dice and keep the lowest.
    """
    mocker.patch('random.randint', side_effect=[15, 3, 6, 18, 10, 1])
    num = 5
    size = 20
    assert op.keep_low_die(num, size) == 3


def test_wild_die(mocker):
    """Given a number of dice and the size of the die, roll that many
    dice and add them together. If the first die rolled doesn't roll
    the highest number possible for that die, it is counted like any
    other die.
    """
    mocker.patch('random.randint', side_effect=[3, 4, 1, 5, 4, 2])
    num = 5
    size = 6
    assert op.wild_die(num, size) == 17


def test_wild_die_explodes(mocker):
    """Given a number of dice and the size of the die, roll that many
    dice and add them together. If the first die rolled rolls the
    highest number possible for that die, it explodes.
    """
    mocker.patch('random.randint', side_effect=[6, 2, 4, 1, 5, 4, 3])
    num = 5
    size = 6
    assert op.wild_die(num, size) == 22


def test_wild_die_is_one(mocker):
    """Given a number of dice and the size of the die, roll that many
    dice and add them together. If the first die rolled rolls the
    a one, the roll is zero.
    """
    mocker.patch('random.randint', side_effect=[1, 4, 1, 5, 4])
    num = 5
    size = 6
    assert op.wild_die(num, size) == 0


# Pool degeneration test cases.
def test_pool_concatenate(mocker):
    """Concatenate the members in the pool."""
    pool = (3, 1, 4)
    assert op.pool_concatenate(pool) == 314


def test_pool_count(mocker):
    """Count the members in the pool."""
    pool = (3, 1, 4)
    assert op.pool_count(pool) == 3


def test_pool_count(mocker):
    """Sum the members in the pool."""
    pool = (3, 1, 4)
    assert op.pool_sum(pool) == 8


def test_count_successes(mocker):
    """Count the number of values above or equal to a target."""
    pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
    target = 5
    assert op.count_successes(pool, target) == 5


def test_count_successes_with_botch(mocker):
    """Count the number of values above or equal to a target and
    remove botches.
    """
    pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
    target = 5
    assert op.count_successes_with_botch(pool, target) == 3


# Pool generation operation test cases.
def test_dice_pool(mocker):
    """Generate a dice pool."""
    pool = (3, 5, 1, 10, 8, 4, 3, 8)
    mocker.patch('random.randint', side_effect=pool)
    num = 7
    size = 10
    assert op.dice_pool(num, size) == pool[:num]


def test_exploding_pool(mocker):
    """Generate a dice pool."""
    rolls = (2, 6, 3, 1, 1, 6, 6, 1, 3)
    mocker.patch('random.randint', side_effect=rolls)
    num = 6
    size = 6
    assert op.exploding_pool(num, size) == (2, 9, 1, 1, 13, 3)


# Pool generation operations using `secrets` test cases.
def test_dice_pool_secrets(mocker):
    """When called, :func:`operator.dice_pool_secrets` should use
    :mod:`secrets` to generate a dice pool.
    """
    pool = (3, 5, 1, 10, 8, 4, 3, 8)
    mocker.patch('secrets.randbelow', side_effect=(n - 1 for n in pool))
    num = 7
    size = 10
    assert op.dice_pool_secrets(num, size) == pool[:num]


# Pool operation test cases.
def test_pool_keep_above():
    """Keep dice equal to or below above the given value from the
    pool.
    """
    pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
    keep = 4
    assert op.pool_keep_above(pool, keep) == (5, 6, 4, 5, 6, 6)


def test_pool_keep_below():
    """Keep dice equal to or below the given value from the pool."""
    pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
    ceiling = 4
    assert op.pool_keep_below(pool, ceiling) == (1, 2, 4, 1, 3)


def test_pool_cap():
    """Dice in the pool are capped at the given value."""
    pool = (1, 2, 3, 10, 5, 6, 7, 8, 9, 4)
    cap = 7
    assert op.pool_cap(pool, cap) == (1, 2, 3, 7, 5, 6, 7, 7, 7, 4)


def test_pool_floor():
    """Dice in the pool are floored at the given value."""
    pool = (1, 2, 3, 10, 5, 6, 7, 8, 9, 4)
    cap = 3
    assert op.pool_floor(pool, cap) == (3, 3, 3, 10, 5, 6, 7, 8, 9, 4)


def test_pool_keep_high():
    """Keep the given number of highest dice from the pool."""
    pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
    keep = 5
    assert op.pool_keep_high(pool, keep) == (5, 6, 5, 6, 6)


def test_pool_keep_low():
    """Keep the given number of highest dice from the pool."""
    pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
    keep = 3
    assert op.pool_keep_low(pool, keep) == (1, 2, 1)


def test_pool_keep_below():
    """Perform a modulo on all members."""
    pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
    divisor = 3
    assert op.pool_modulo(pool, divisor) == (1, 2, 2, 0, 1, 2, 1, 0, 0, 0)


def test_pool_keep_below():
    """Keep dice equal to or below the given value from the pool."""
    pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
    cut = 5
    assert op.pool_remove(pool, cut) == (1, 2, 6, 4, 1, 6, 3, 6)



"""
parse
~~~~~

Parse dice notation.
"""
import operator
from collections.abc import Callable, Sequence
from functools import wraps
from typing import Any, Optional

from yadr import operator as yo
from yadr.model import (
    CompoundResult,
    DiceMapping,
    Result,
    Token,
    TokenInfo,
    id_tokens,
    op_tokens
)


# The dice map.
# This needs to not be a global value, but it will require a very large
# change to this module to get that to work. This will work for now.
dice_map: dict[str, DiceMapping] = {}


# Parser specific operations.
def map_result(
    result: int | tuple[int, ...],
    key: str
) -> str | int | tuple[str, ...]:
    """Map a roll result to a dice map."""
    if isinstance(result, int):
        return dice_map[key][result]
    new_result = [map_result(n, key) for n in result]
    str_result = tuple(str(item) for item in new_result)
    return str_result


# Exceptons
class IsMap(Exception):
    """Raised to tell the parser not to expect results from the
    current roll because the roll is a dice mapping.
    """


class NoResult(Exception):
    """Raised to tell the parser no result was returned for the
    passed :ref:`YADN` tokens.
    """


# Utility classes and functions.
def _bad_type(side: str, need: str, got: Any) -> None:
    """Raise a TypeError."""
    name = type(got).__name__
    msg = f'{side.title()} operand must be {need}. Was {name}.'
    raise ValueError(msg)


class Tree:
    """A binary tree used to execute parsed :ref:`YADN`.

    :param kind: The type of token to be executed.
    :param value: The specific value of the token for execution.
    :param left: (Optional.) The branch of tokens to be executed first.
    :param right: (Optional.) The branch of tokens to be executed last.
    :param dice_map: (Optional.) Dice maps to use during execution.
    :return: None.
    :rtype: :class:`NoneType`
    """
    def __init__(
        self,
        kind: Token,
        value: Result,
        left: Optional['Tree'] = None,
        right: Optional['Tree'] = None,
        dice_map: Optional[dict[str, DiceMapping]] = None
    ) -> None:
        self.kind = kind
        self.value = value
        self.left = left
        self.right = right
        if dice_map is None:
            dice_map = {}
        self.dice_map = dice_map

    def __repr__(self):
        name = self.__class__.__name__
        return f'{name}(kind={self.kind}, value={self.value})'

    def compute(self):
        """Execute the tree.

        :return: The result of the identity or operation token.
        :rtype: Result

        .. warning::
            The return here is too complex to type check. At least
            it's too complex for me to type check at this point. I'll
            keep trying, but for now it's not annotated.
        """
        # Simple identities do not need any computation.
        if self.kind in id_tokens:
            return self.value

        # The only tokens that should make it into Trees are identity
        # or operation tokens. If a different type of token is here,
        # there is a problem that can't be easily resolved.
        if self.kind not in op_tokens:
            msg = f'Unknown token {self.kind}'
            raise TypeError(msg)

        # Compute the value of the tokens that are higher in the
        # order of operations than this tree and on the left-side
        # of this expression.
        left = self.left.compute()

        # Compute the value of the tokens that are higher in the
        # order of operations than this tree and on the right-side
        # of this expression.
        right = self.right.compute()

        # Determine the operation to run.
        try:
            op = yo.ops[self.value]
        except KeyError:
            if self.value != 'm':
                msg = f'Operator not recognized: {self.value}.'
                raise ValueError(msg)
            op = self._map_result

        # Perform the operation on the left and right values,
        # returning the result.
        return op(left, right)

    def _map_result(
        self, result: int | tuple[int, ...],
        key: str
    ) -> str | int | tuple[str, ...]:
        """Map a roll result to a dice map."""
        if isinstance(result, int):
            return self.dice_map[key][result]
        new_result = [self._map_result(n, key) for n in result]
        str_result = tuple(str(item) for item in new_result)
        return str_result


class Unary(Tree):
    """A unary tree used to execute parsed :ref:`YADN`.

    :param kind: The type of token to be executed.
    :param value: The specific value of the token for execution.
    :param child: (Optional.) The branch of tokens to be executed first.
    :return: None.
    :rtype: :class:`NoneType`
    """
    def __init__(
        self,
        kind: Token,
        value: Result,
        child: Optional['Tree'] = None
    ) -> None:
        self.kind = kind
        self.value = value
        self.child = child

    def compute(self):
        """Execute the tree.

        :return: The result of the identity or operation token.
        :rtype: Result

        .. warning::
            This is not type annotated to remain consistent with
            :meth:`yadr.parser.Tree.compute`. See the docstring there
            for why.
        """
        # Simple identities do not need any computation.
        if self.kind in id_tokens:
            return self.value

        # Compute the result of any children.
        child = self.child.compute()

        # Determine the operation to perform.
        if self.kind in op_tokens:
            op = yo.ops[self.value]

        # Perform the operation and return the result.
        return op(child)


# Parser class.
class Parser:
    """A state machine for parsing :ref:`YADN`.

    How Parsing Works
    =================
    Essentially, parsing turns an ordered list of tokens into a
    tree structure for execution. For example, let's say we have
    the :ref:`YADN` expression::

        3 * ( 4 - 2 )

    That is lexed into the tokens::

        Token(NUMBER, 3)
        Token(MD_OPERATOR, '*')
        Token(GROUP_OPEN, '(')
        Token(NUMBER, 4)
        Token(AS_OPERATOR, '-')
        Token(NUMBER, 2)
        Token(GROUP_CLOSE, ')')

    Which is then parsed into the tree::

            *
           / \\
          -   3
         / \\
        4   2

    Where each token ends up in each tree is dependent on the
    :ref:`ops_order` defined by :ref:`YADN`.

    """
    def __init__(self) -> None:
        self.dice_map: dict[str, DiceMapping] = dict()
        self.top_rule = self._map_operator

    # Public method.
    def parse(
        self,
        tokens: Sequence[TokenInfo]
    ) -> None | Result | CompoundResult:
        """Parse one or more die rolls.

        :param tokens: A sequence of lexed :ref:`YADN` tokens to parse.
        :return: A class defined in either :class:`yadr.model.Result` or
            :class:`yadr.model.CompoundResult`.
        :rtype: Result | CompoundResult
        """
        # Split tokens into rolls for parsing.
        rolls = []
        while (Token.ROLL_DELIMITER, ';') in tokens:
            index = tokens.index((Token.ROLL_DELIMITER, ';'))
            roll = tokens[0:index]
            rolls.append(roll)
            tokens = tokens[index + 1:]
        else:
            rolls.append(tokens)

        # Parse each roll.
        results: list[Result] = []
        for roll in rolls:
            try:
                results.append(self._parse_roll(roll))
            except IsMap:
                continue
            except NoResult:
                continue

        # Return the results of the rolls.
        if len(results) > 1:
            return CompoundResult(results)
        elif results:
            return results[0]
        return None

    def _make_tree(self, kind: Token, value: Result) -> Tree:
        """Tranform tokens into trees for execution."""
        return Tree(kind, value, dice_map=self.dice_map)

    def _parse_roll(self, tokens: Sequence[TokenInfo]) -> Result:
        """Parse a sequence of YADN tokens."""
        # Parse the tokens into a tree.
        trees = [self._make_tree(kind, value) for kind, value in tokens]
        trees = trees[::-1]
        parsed = self.top_rule(trees)

        # Execute the parsed tree.
        if not parsed:
            raise NoResult('The parsed string did not create a Tree.')
        return parsed.compute()

    # Parsing rules.
    def _identity(self, trees: list[Tree]) -> Tree:
        """Parse an identity."""
        identity_tokens = [
            Token.BOOLEAN,
            Token.NUMBER,
            Token.POOL,
            Token.QUALIFIER,
        ]
        tree = trees.pop()

        # If the tree is an identity, return it.
        if tree.kind in identity_tokens:
            return tree

        # If the tree is a dice map, add it to the dice map and raise
        # IsMap to prevent the parse from trying to return it as a
        # result. The asserts in here are to keep mypy happy. They
        # probably mean I need to rework how dice maps are handled to
        # keep them out of this parser.
        elif tree.kind == Token.MAP:
            assert isinstance(tree.value, tuple)
            name, map_ = tree.value
            assert isinstance(name, str)
            assert isinstance(map_, dict)
            self.dice_map[name] = map_
            raise IsMap('Roll is a map.')

        # If the tree is starting a group, parse through that group.
        elif tree.kind == Token.GROUP_OPEN:
            expression = self.top_rule(trees)
            if trees[-1].kind == Token.GROUP_CLOSE:
                _ = trees.pop()
            return expression

        # Handle anything unexpected that got here.
        else:
            msg = f'Unrecognized token {tree.kind}'
            raise TypeError(msg)

    def _pool_gen_operator(self, trees: list[Tree]) -> Tree:
        """Parse pool generation operator."""
        rule = self._binary_operator
        rule_affects = Token.POOL_GEN_OPERATOR
        next_rule = self._identity
        return rule(rule_affects, next_rule, trees)

    def _pool_operator(self, trees: list[Tree]) -> Tree:
        """Parse pool operator."""
        rule = self._binary_operator
        rule_affects = Token.POOL_OPERATOR
        next_rule = self._pool_gen_operator
        return rule(rule_affects, next_rule, trees)

    def _u_pool_degen_operator(self, trees: list[Tree]) -> Tree:
        """Parse unary pool degeneration."""
        rule = self._unary_operator
        rule_affects = Token.U_POOL_DEGEN_OPERATOR
        next_rule = self._pool_operator
        return rule(rule_affects, next_rule, trees)

    def _pool_degen_operator(self, trees: list[Tree]) -> Tree:
        """Parse unary pool degeneration."""
        rule = self._binary_operator
        rule_affects = Token.POOL_DEGEN_OPERATOR
        next_rule = self._u_pool_degen_operator
        return rule(rule_affects, next_rule, trees)

    def _dice_operator(self, trees: list[Tree]) -> Tree:
        """Parse dice operators."""
        rule = self._binary_operator
        rule_affects = Token.DICE_OPERATOR
        next_rule = self._pool_degen_operator
        return rule(rule_affects, next_rule, trees)

    def _ex_operator(self, trees: list[Tree]) -> Tree:
        """Parse exponentiation."""
        rule = self._binary_operator
        rule_affects = Token.EX_OPERATOR
        next_rule = self._dice_operator
        return rule(rule_affects, next_rule, trees)

    def _md_operator(self, trees: list[Tree]) -> Tree:
        """Parse addition and subtraction."""
        rule = self._binary_operator
        rule_affects = Token.MD_OPERATOR
        next_rule = self._ex_operator
        return rule(rule_affects, next_rule, trees)

    def _as_operator(self, trees: list[Tree]) -> Tree:
        """Parse addition and subtraction."""
        rule = self._binary_operator
        rule_affects = Token.AS_OPERATOR
        next_rule = self._md_operator
        return rule(rule_affects, next_rule, trees)

    def _comparison_operator(self, trees: list[Tree]) -> Tree:
        """Parse comparisons."""
        rule = self._binary_operator
        rule_affects = Token.COMPARISON_OPERATOR
        next_rule = self._as_operator
        return rule(rule_affects, next_rule, trees)

    def _options_operator(self, trees: list[Tree]) -> Tree:
        """Parse comparisons."""
        rule = self._binary_operator
        rule_affects = Token.OPTIONS_OPERATOR
        next_rule = self._comparison_operator
        return rule(rule_affects, next_rule, trees)

    def _choice_operator(self, trees: list[Tree]) -> Tree:
        """Parse chocies."""
        rule = self._binary_operator
        rule_affects = Token.CHOICE_OPERATOR
        next_rule = self._options_operator
        return rule(rule_affects, next_rule, trees)

    def _map_operator(self, trees: list[Tree]) -> Tree:
        """Parse dice mapping operators."""
        rule = self._binary_operator
        rule_affects = Token.MAPPING_OPERATOR
        next_rule = self._choice_operator
        return rule(rule_affects, next_rule, trees)

    # Base rules.
    def _binary_operator(
        self, rule_affects: Token,
        next_rule: Callable[[list[Tree]], Tree],
        trees: list[Tree]
    ) -> Tree:
        """Parse a binary operator."""
        left = next_rule(trees)
        while trees and trees[-1].kind == rule_affects:
            tree = trees.pop()
            tree.left = left
            tree.right = next_rule(trees)
            left = tree
        return left

    def _unary_operator(
        self, rule_affects: Token,
        next_rule: Callable[[list[Tree]], Tree],
        trees: list[Tree]
    ) -> Tree:
        """Parse an unary operator."""
        if trees[-1].kind == rule_affects:
            tree = trees.pop()
            unary = Unary(tree.kind, tree.value)
            unary.child = next_rule(trees)
            return unary
        return next_rule(trees)
